-------------------------------------------------------------------------------
-- Localization libraries
-------------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("SimpleSelfRebuff")
local BS 

-------------------------------------------------------------------------------
-- Aliases & constants
-------------------------------------------------------------------------------

local activationCheckEvents = {
	'PLAYER_DEAD', 'PLAYER_UNGHOST', 'PLAYER_ALIVE',
	'PLAYER_UPDATE_RESTING', 'PLAYER_AURAS_CHANGED',
	'PLAYER_CONTROL_GAINED', 'PLAYER_CONTROL_LOST',
}

local CATEGORY_TRACKING = L["Tracking"] or "Tracking"
local CATEGORY_MAINHAND = L["Main weapon"] or "Main weapon"
local CATEGORY_OFFHAND  = L["Off-hand weapon"] or "Off-hand weapon"

local SOURCE_SPELL = 'spell'
local SOURCE_ITEM  = 'item'

local TARGET_AURA     = 'aura'
local TARGET_TRACKING = 'tracking'
local TARGET_MAINHAND = 'main-hand'
local TARGET_OFFHAND  = 'off-hand'

-------------------------------------------------------------------------------
-- Locals
-------------------------------------------------------------------------------

local categories = {}
local sources = {}
local targets = {}

local buffToCast

local monitoringActive

local db, db_char

-------------------------------------------------------------------------------
-- Externals
-------------------------------------------------------------------------------

local Waterfall
 
local AceOO = AceLibrary('AceOO-2.0')
local Class = AceOO.Class

-------------------------------------------------------------------------------
-- Addon declaration
-------------------------------------------------------------------------------

SimpleSelfRebuff = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceDB-2.0", "AceModuleCore-2.0")
local SimpleSelfRebuff = SimpleSelfRebuff

-- Static early initialization
do
	local self = SimpleSelfRebuff

	self:RegisterDB("SimpleSelfRebuffDB", "SimpleSelfRebuffDBPC")

	self.options = {	
		type = 'group',
		args = {
			menu = {
				type = 'execute',
				name = L["Menu"],
				desc = L["Menu"],
				func = function() Waterfall:Open('SimpleSelfRebuff') end,
				hidden = function() return not Waterfall end,
				wfHidden = true,
				order = 50,
			},
			menuHeader = {
				type = 'header',
				name = L["Global options"],
				order = 51,
				wfHidden = true,					
				hidden = function() return not Waterfall end,
			},		
			--
			rebuffThreshold = {
				type = 'range',
				name = L["Rebuff time threshold"],
				desc = L["Any buff with remaining time less than this value will be recast. Select 0 to disable."],
				min = 0,
				max = 180,
				step = 1,
				bigStep = 5,
				get = function() return db.rebuffThreshold end,
				set = function(value)
					db.rebuffThreshold = value
					self:CheckRebuff()
				end,
			},
			--
			buffHeader = {
				type = 'header',
				name = L["Buff selection"],
				order = 199,
				wfHidden = true,
			},
			--
			diag = {
				type = 'execute',
				name = 'Diagnostic',
				desc = 'Dump some data to help bug fixing. Please includes this output into your bug reports.',
				func = 'DumpDiagnostic',
				order = -3.5,
			},
		}
	}
		
	self.CATEGORY_TRACKING = CATEGORY_TRACKING
	self.CATEGORY_MAINHAND = CATEGORY_MAINHAND 
	self.CATEGORY_OFFHAND  = CATEGORY_OFFHAND 
	
	self.SOURCE_SPELL = SOURCE_SPELL
	self.SOURCE_ITEM  = SOURCE_ITEM
	
	self.TARGET_AURA     = TARGET_AURA
	self.TARGET_TRACKING = TARGET_TRACKING 
	self.TARGET_MAINHAND = TARGET_MAINHAND
	self.TARGET_OFFHAND  = TARGET_OFFHAND

	-- For debugging
	SSR = self
	
	self.categories = categories
	self.sources = sources
	self.targets = targets
end

-------------------------------------------------------------------------------
-- Helpers
-------------------------------------------------------------------------------

local function error(msg, ...)
	if select('#', ...) > 0 then
		msg = msg:format(...)
	end
	return _G.error(msg, 2)
end

local function erase(...)
	for i=1,select('#', ...) do
		local t = select(i, ...)
		if type(t) == 'table' then
			for k in pairs(t) do
				t[k] = nil 
			end
		end
	end
end

local function loadHash(t, ...)
	for i = 1, select('#', ...), 2 do
		local k,v = select(i, ...)
		if t[k] == nil then
			t[k] = v
		end
	end
end

local function colorize(text, r, g, b)
	return ("|cff%02x%02x%02x%s|r"):format(math.floor(r*255), math.floor(g*255), math.floor(b*255), text)
end
SimpleSelfRebuff.colorize = colorize

local new, del
do
	local list = setmetatable({}, {__mode='k'})
	
	function new(...)
		local t = next(list)
		if t then
			list[t] = nil
			for i = 1, select('#', ...) do
				t[i] = select(i, ...)
			end
			return t
		else
			return {...}
		end
	end
	function del(t)
		if type(t) == 'table' then
			setmetatable(t, nil)
			for k in pairs(t) do
				t[k] = nil
			end
			list[t] = true
		end
		return nil
	end
end
SimpleSelfRebuff.new = new
SimpleSelfRebuff.del = del

local function formatDuration(duration, default)
	if not duration or duration < 1 then
		return default or "-"
	elseif duration >= 3600 then
		return HOUR_ONELETTER_ABBR:format(math.ceil(duration/3600))
	elseif duration >= 60 then
		return MINUTE_ONELETTER_ABBR:format(math.ceil(duration/60))
	else
		return SECOND_ONELETTER_ABBR:format(duration)
	end
end
SimpleSelfRebuff.formatDuration = formatDuration

-------------------------------------------------------------------------------
-- Initializing
-------------------------------------------------------------------------------

function SimpleSelfRebuff:_rev(tag)
	self.revision = math.max(self.revision or 0, tonumber(tag:match("%d+")))
end

-- Addon setup
function SimpleSelfRebuff:OnInitialize()
	self:_rev("$Revision: 43976 $")
	self.version = "r" .. SimpleSelfRebuff.revision
	self.date = string.sub("$Date: 2007-07-17 02:29:18 -0400 (Tue, 17 Jul 2007) $", 8, 17)

	self:RegisterDefaults('profile', {
		rebuffThreshold = 30,
	})

	self:RegisterDefaults('char', {
		categories = {},
	})
			
	db = self.db.profile
	db_char = self.db.char

	self:RegisterChatCommand({"/ssr", "/simpleselfrebuff"}, self.options)
	
	local AceDebug = AceLibrary:HasInstance('AceDebug-2.0') and AceLibrary('AceDebug-2.0')
	if AceDebug then
		AceDebug:embed(self)
		local opt = AceDebug:GetAceOptionsDataTable(self)
		self.options.args.debug = opt.debug
		SimpleSelfRebuff:SetDebugging(UnitName('player') == 'Adirelle')
	else
		function self.Debug() end
	end
	self.modulePrototype.Debug = function(module, ...) self:Debug(...) end
	
	Waterfall = AceLibrary:HasInstance('Waterfall-1.0') and AceLibrary('Waterfall-1.0')
	if Waterfall then
		Waterfall:Register('SimpleSelfRebuff', 
			'aceOptions', self.options,
			'title','SimpleSelfRebuff',
			'colorR',0.8,
			'colorG',0.2,
			'colorB',0.8
		)
	end
		
end

function SimpleSelfRebuff:OnEnable(first)
	if first then
		self:ProcessRegistry()
	end

	self:RegisterBucketEvent(activationCheckEvents, 0, 'CheckActivation')
	self:RegisterBucketEvent('SimpleSelfRebuff_BuffChanged', 0, 'CheckRebuff')
	self:RegisterEvent('SimpleSelfRebuff_AvailableBuffsChanged', 'AvailableBuffsChanged')

	self:CheckActivation()
end

function SimpleSelfRebuff:OnDisable(first)
	self:SetMonitoringActive(false)
end

function SimpleSelfRebuff:AvailableBuffsChanged()
	self:UpdateCategories()
	self:UpdateOptions()
	self:CheckActivation()
end

function SimpleSelfRebuff:UpdateCategories()
	for name, category in pairs(categories) do
		category:Update()
	end
end

function SimpleSelfRebuff:DumpDiagnostic()
	self:Print("Version %s", self.version)
	self:Print("=== Categories ===")
	for categoryName, category in pairs(categories) do
		self:Print("* %s: selected=%q, actual=%q", 
			categoryName, 
			category:GetExpectedBuff() or "none",
			category:GetActualBuff() or "none"
		)
	end
	self:Print("=== Buff targets ===")
	for name,target in pairs(targets) do
		if type(target.DumpDiagnostic) == "function" then
			self:Print("* %s:", name)
			target:DumpDiagnostic()
		end
	end
end

-------------------------------------------------------------------------------
-- Options handling
-------------------------------------------------------------------------------

do
	local NO_BUFF = L["None"]
	
	local function getSelectCategoryBuff(category)
		local buff = category:GetExpectedBuff()
		return buff and buff.name or NO_BUFF
	end
	
	local function getToggleCategoryBuff(category)
		local buff = category:GetExpectedBuff()
		return buff and (buff.name == category.name)
	end
	
	local function setCategoryBuff(category, name)
		if name == NO_BUFF then
			name = nil
		elseif name == true then
			name = category.name
		end
		category:SetExpectedBuff(name)
	end
	
	function SimpleSelfRebuff:BuildOptions()
		local opts = self.options.args
		for categoryName, category in pairs(categories) do
			local categoryName, category = categoryName, category
			local option = {
				name = categoryName,
				desc = categoryName,
				passValue = category,
				set = setCategoryBuff,
				hidden = function() return category.found == 0 end,
				disabled = function() return category.castable == 0 end,
				order = 200,
			}
			if category.count > 1 then
				option.type = 'text'
				option.get = getSelectCategoryBuff
				option.validate = { NO_BUFF }
			else
				option.type = 'toggle'
				option.get = getToggleCategoryBuff
			end
			opts[categoryName] = option
		end
		self:UpdateOptions()
	end
	
	function SimpleSelfRebuff:UpdateOptions()
		for categoryName, category in pairs(categories) do
			if category.count > 1 then
				local validate = self.options.args[category.name].validate
				erase(validate)
				for name,buff in pairs(category.buffs) do
					if buff:IsCastable() then
						tinsert(validate, buff.name)
					end
				end
				table.sort(validate)				
				tinsert(validate, 1, NO_BUFF)
			end
		end
	end
end

-------------------------------------------------------------------------------
-- Activation
-------------------------------------------------------------------------------

function SimpleSelfRebuff:SetMonitoringActive(value)
	if value and not monitoringActive then
		monitoringActive = true
		self:Debug('Activated')
		
		self:ScheduleRepeatingEvent('SimpleSelfRebuff_BuffChanged', 1.5)
		
		self:TriggerEvent('SimpleSelfRebuff_MonitoringEnabled')
		self:TriggerEvent('SimpleSelfRebuff_UpdateDisplay')
		
	elseif not value and monitoringActive then
		monitoringActive = nil
		self:Debug('Disactivated')

		self:CancelScheduledEvent('SimpleSelfRebuff_BuffChanged')
		
		self:TriggerEvent('SimpleSelfRebuff_MonitoringDisabled')		
		self:TriggerEvent('SimpleSelfRebuff_UpdateDisplay')
	end
end

function SimpleSelfRebuff:CheckActivation()
	self:SetMonitoringActive(
		not (UnitIsDeadOrGhost('player') or IsResting())
	)
end

function SimpleSelfRebuff:IsMonitoringActive()
	return monitoringActive
end

-------------------------------------------------------------------------------
-- Core
-------------------------------------------------------------------------------

function SimpleSelfRebuff:CheckRebuff()

	-- Avoid to erase pending cast
	if buffToCast and buffToCast:IsUsable() and buffToCast:IsExpectedBuff() and not buffToCast.category:GetActualBuff() then
		return				
	end

	-- Determine if another buff needs refresh
	
	local rebuffThreshold = db.rebuffThreshold > 0 and db.rebuffThreshold or nil
	local newBuffToCast, buffTimeLeft, shortestBuff, shortestBuffDuration
	
	for categoryName,category in pairs(categories) do
		local expectedBuff = category:GetExpectedBuff()
		if expectedBuff and expectedBuff:IsUsable() then			
			local actualBuff, timeLeft = category:GetActualBuff()
			
			-- Check if it is missing
			if not actualBuff then
				newBuffToCast = expectedBuff
				--self:Debug('Missing buff: %q', expectedBuff)
				break
				
			-- Check rebuffing threshold
			elseif rebuffThreshold and timeLeft and actualBuff == expectedBuff then
				if not shortestBuffDuration or timeLeft < shortestBuffDuration then
					shortestBuffDuration, shortestBuff = timeLeft, actualBuff
				end
			end
				
		end
	end
	
	if not newBuffToCast and shortestBuff and shortestBuffDuration < rebuffThreshold then
		newBuffToCast = shortestBuff
		buffTimeLeft = shortestBuffDuration
		self:Debug('Buff needs refresh: %q (%q)', newBuffToCast, buffTimeLeft)
	end

	if newBuffToCast ~= buffToCast then
		buffToCast = newBuffToCast
		self:TriggerEvent('SimpleSelfRebuff_PendingCastChanged', buffToCast, buffTimeLeft)
		self:TriggerEvent('SimpleSelfRebuff_UpdateDisplay')
	end

end

function SimpleSelfRebuff:GetPendingCast()
	return buffToCast
end

-------------------------------------------------------------------------------
-- Module prototype
-------------------------------------------------------------------------------

SimpleSelfRebuff:SetModuleMixins('AceEvent-2.0')
SimpleSelfRebuff.modulePrototype.core = SimpleSelfRebuff

function SimpleSelfRebuff:OnModuleCreate(name, module)
	self:SetModuleDefaultState(module, true)
end

function SimpleSelfRebuff.modulePrototype:OnInitialize()
	self.db = self.core:AcquireDBNamespace(self.name)
	
	if type(self.OnModuleInitialize) == "function" then
		self:OnModuleInitialize()
	end
	
	local options = self.options or {}
	
	options.enable = {
		type = 'toggle',
		name = L['Enable'],
		desc = L['Enable %s.']:format(self.optionName or self.name),
		order = 1,
		handler = self.core,
		passValue = self,
		get = "IsModuleActive",
		set = "ToggleModuleActive",
	}
	
	self.core.options.args[self.name] = {
		type = 'group',
		name = self.optionName or self.name,
		desc = self.optionDesc or self.name,
		handler = self,
		args = options,
		order = self.optionOrder,
	}	
	
end

function SimpleSelfRebuff.modulePrototype:OnEnable(first)
	if type(self.OnModuleEnable) == "function" then
		self:OnModuleEnable(first)
	end
end

function SimpleSelfRebuff.modulePrototype:OnDisable()
	if type(self.OnModuleDisable) == "function" then
		self:OnModuleDisable()
	end
end

function SimpleSelfRebuff.modulePrototype:RegisterDefaults(kind, defaults)
	self.core:RegisterDefaults(self.name, kind, defaults)
end


-------------------------------------------------------------------------------
-- Base buff class
-------------------------------------------------------------------------------
local CategoryClass, BuffClass

BuffClass = Class()

function BuffClass.prototype:init(name, category, source, target, ...)
	if type(name) ~= 'string' then
		error('Argument #2 to BuffClass:new() should be a string, not %q', type(name))
	end		
	
	loadHash(self, ...)
	self.name = name
	self.category = category
	self.source = source
	self.target = target
	
	BuffClass.super.prototype.init(self)

	source:RegisterBuff(self)
	target:RegisterBuff(self)	
end

function BuffClass.prototype:ChangeName(name)
	if type(name) ~= 'string' then
		error(2, "Argument #2 to SetName should be a string, not %q", type(name))
	elseif name == self.name then
		return
	end
	
	self.category.buffs[self.name] = nil
	self.target.buffs[self.name] = nil
	self.source.buffs[self.name] = nil
	
	self.name = name
	
	self.category.buffs[name] = self
	self.target.buffs[name] = self
	self.source.buffs[name] = self
end

function BuffClass.prototype:IsCastable()
	if self.found then
		if self.minLevel and UnitLevel('player') < self.minLevel then
			return false
		elseif self.maxLevel and UnitLevel('player') > self.maxLevel then
			return false
		elseif type(self.target.IsTargetValid) == "function" then
			return self.target:IsTargetValid(self)
		end
		return true
	end
	return false
end

function BuffClass.prototype:SetAsActualBuff(flag, ownBuff, timeLeft)
	if flag then
		self.category:SetActualBuff(self, ownBuff, timeLeft)
	else
		self.category:SetActualBuff(nil, nil, nil)
	end
end

function BuffClass.prototype:IsActualBuff()
	return self.category:GetActualBuff() == self
end

function BuffClass.prototype:IsExpectedBuff()
	return self.category:GetExpectedBuff() == self
end

function BuffClass.prototype:IsUsable()
	return self.source:IsBuffUsable(self)
end

function BuffClass.prototype:IsInCooldown()
	return self.source:IsBuffInCooldown(self)
end

function BuffClass.prototype:SetupSecureButton(button)
	self.source:SetupSecureButton(self, button)
	self.target:SetupSecureButton(self, button)
end

function BuffClass.prototype:ToString()
	return '<Buff-"'..self.name..'">'
end

-------------------------------------------------------------------------------
-- Category class
-------------------------------------------------------------------------------

CategoryClass = Class()

function CategoryClass.prototype:init(name, target, source, ...)
	if type(name) ~= 'string' then
		error('Argument #2 to CategoryClass:new() should be a string, not %q', type(name))
	end
	CategoryClass.super.prototype.init(self)
	self.name = name
	self.buffs = {}
	self.count = 0
	self.found = 0
	self.target = target or TARGET_AURA
	self.source = source or SOURCE_SPELL
	loadHash(self, ...)
end

function CategoryClass.prototype:Update()

	self.count = 0
	self.found = 0
	self.castable = 0

	for name,buff in pairs(self.buffs) do
		self.count = self.count + 1
		if buff.found then
			self.found = self.found + 1
			if buff:IsCastable() then
				self.castable = self.castable + 1
			end
		end
		if buff.name ~= name then
			self.buffs[name] = nil
			self.buffs[buff.name] = buff
		end		
	end
	
	local expected = self:GetExpectedBuff()
	if expected and not expected.found then
		self:SetExpectedBuff(nil)
	end
	
end

function CategoryClass.prototype:add(name, ...)

	if self.buffs[name] then
		error("%q already registered in category %q", name, self.name)
	end
	
	local source = self.source and sources[self.source]
	local target = self.target and targets[self.target]
	
	if not source then
		error("unknown source for %q: %q", name, self.source)
	elseif not target then
		error("unknown target for %q: %q", name, self.target)
	end		
		
	local buff = BuffClass:new(name, self, source, target, ...)
	self.buffs[buff.name] = buff
	self.count = self.count + 1

	return self
end

function CategoryClass.prototype:addMulti(...)
	for i=1,select('#', ...) do
		local name = select(i, ...)
		self:add(name)
	end
	return self
end

function CategoryClass.prototype:GetExpectedBuff()
	local name = db_char.categories[self.name]
	if self.buffs[name] then
		return self.buffs[name]
	else
		db_char.categories[self.name] = nil
	end
end

function CategoryClass.prototype:SetExpectedBuff(buff)
	local buffName 
	if type(buff) == 'string' then
		buffName = buff
	elseif buff then
		buffName = buff.name
	end
	if buffName then
		if not self.buffs[buffName] then
			error("Unknown buff %q in category %q", buffName, self.name)
		elseif not self.buffs[buffName].found then
			buffName = nil
		end	
	end
	if buffName ~= db_char.categories[self.name] then
		db_char.categories[self.name] = buffName
		SimpleSelfRebuff:TriggerEvent('SimpleSelfRebuff_BuffSetupChanged', buffName, self.name)
	end
end

function CategoryClass.prototype:GetActualBuff()
	local timeLeft = self.actualBuffExpiration and math.floor(self.actualBuffExpiration - GetTime())
	if not timeLeft or timeLeft <= 0 then
		timeLeft = nil
	end
	return self.actualBuff, timeLeft
end

function CategoryClass.prototype:SetActualBuff(current, ownBuff, timeLeft)
	if current and not ownBuff then
		local expected = self:GetExpectedBuff()
		if expected and (expected.subcat or current.subcat) and expected.subcat ~= current.subcat then
			return
		end	
	end	
	local expiration = (current and type(timeLeft) == "number" and math.floor(GetTime() + timeLeft)) or nil
	local dirty = (current ~= self.actualBuff) or (expiration ~= self.actualBuffExpiration)
	self.actualBuff = current
	self.actualBuffExpiration = expiration
	if dirty then
		SimpleSelfRebuff:TriggerEvent('SimpleSelfRebuff_BuffChanged')
		return true
	end
end

function CategoryClass.prototype:ToString()
	return '<Category-"'..self.name..'">'
end

-------------------------------------------------------------------------------
-- Some mixin helpers
-------------------------------------------------------------------------------

local BuffAspectClass, BuffSourceClass, BuffTargetClass

do
	
	-----------------------------------------------------------------------------
	-- Abstract buff aspect class 
	-----------------------------------------------------------------------------
	
	BuffAspectClass = Class(nil, "AceEvent-2.0")
	
	BuffAspectClass.virtual = true
	
	function BuffAspectClass.prototype:init()
		BuffAspectClass.super.prototype.init(self)
		self.allBuffs = {}
		self.buffs = {}
		self:RegisterMonitoringEvents()
		self:RegisterEvent('SimpleSelfRebuff_RegistryProcessed', 'OnRegistryProcessed', true)
	end
	
	function BuffAspectClass.prototype:Debug(...)
		SimpleSelfRebuff:Debug(...)
	end
	
	function BuffAspectClass.prototype:_delegate(methodName, ...)
		if type(self[methodName]) == "function" then
			local success, msg = pcall(self[methodName], self, ...)
			if success then
				return msg
			else
				geterrorhandler()(msg)
		end		
	end
	end
	
	function BuffAspectClass.prototype:RegisterMonitoringEvents()
		self:RegisterEvent('SimpleSelfRebuff_MonitoringEnabled', 'OnMonitoringEnable')
		self:RegisterEvent('SimpleSelfRebuff_MonitoringDisabled', 'OnMonitoringDisable')	
	end
	
	function BuffAspectClass.prototype:OnRegistryProcessed()
		self:_delegate('OnInitialize')
	end
	
	function BuffAspectClass.prototype:OnMonitoringEnable()
		if not self._enabled then
			--self:Debug('Enabling %q', self)
			self:_delegate('OnEnable')
			self._enabled = true
		end	
	end

	function BuffAspectClass.prototype:OnMonitoringDisable()
		if self._enabled then
			--self:Debug('Disabling %q', self)
			self:UnregisterAllEvents()
			self:_delegate('OnDisable')
			self._enabled = nil
			self:RegisterMonitoringEvents()
		end
	end
	
	function BuffAspectClass.prototype:RegisterBuff(buff)
		self:_delegate('OnBuffRegister', buff)	
		self.buffs[buff.name] = buff
		self.allBuffs[buff] = true
	end
	
	-----------------------------------------------------------------------------
	-- Source buff aspect class
	-----------------------------------------------------------------------------
	
	BuffSourceClass = Class(BuffAspectClass)
	
	function BuffSourceClass.prototype:init(source, buffUsableCheck, getBuffCooldown)
		BuffSourceClass.super.prototype.init(self)
		self.source = source
		sources[source] = self
		self.buffUsableCheck = buffUsableCheck
		self.getBuffCooldown = getBuffCooldown
	end
	
	function BuffSourceClass.prototype:IsBuffUsable(buff)
		return self.buffUsableCheck(buff.name) and not self:IsBuffInLongCooldown(buff)
	end
	
	function BuffSourceClass.prototype:IsBuffInCooldown(buff)
		local start, duration = self.getBuffCooldown(buff.name)
		local endTime = start and duration and start > 0 and duration > 0 and start + duration
		if endTime and endTime > GetTime() then 
			return true, GetTime()-endTime
		else
			return false
		end
	end

	function BuffSourceClass.prototype:IsBuffInLongCooldown(buff)
		local start, duration = self.getBuffCooldown(buff.name)
		return start > 0 and duration > 1.5 and start+duration > GetTime()
	end
	
	function BuffSourceClass.prototype:ToString()
		return '<BuffSource-'..self.source..'>'
	end

	-----------------------------------------------------------------------------
	-- Target buff aspect class
	-----------------------------------------------------------------------------

	BuffTargetClass = Class(BuffAspectClass)
	
	function BuffTargetClass.prototype:init(target)
		BuffTargetClass.super.prototype.init(self)
		self.target = target
		targets[target] = self
	end
	
	function BuffTargetClass.prototype:ToString()
		return '<BuffTarget-'..self.target..'>'
	end
	
end

-------------------------------------------------------------------------------
-- Buffs from items
-------------------------------------------------------------------------------

do
		
	local ItemBuff = BuffSourceClass:new(SOURCE_ITEM, IsItemUsable, GetItemCooldown)
	local missingItems
		
	function ItemBuff:OnBuffRegister(buff)
		if not buff.itemid then
			error("itemid is missing for ItemBuff %q", buff.name)
		end
		self:ResolveBuffItemId(buff)
	end

	function ItemBuff:ResolveMissingItems()
		if not missingItems then
			return
		end
		self:Debug('Scanning missing items')
		local dirty
		
		for buff in pairs(missingItems) do
			local found, linked, changed = self:ResolveBuffItemId(buff, true) 
			if found and changed then
				dirty = true
			end
			if linked then
				break
			end
		end
		
		if missingItems then
			self:Debug('Planning scanning event')
			self:ScheduleEvent('SimpleSelfRebuff_ResolveItemId', self.ResolveMissingItems, 0.5, self)
		elseif dirty then
			self:TriggerEvent('SimpleSelfRebuff_AvailableBuffsChanged')
		end
	end
	
	ItemBuff.OnInitialize = ItemBuff.ResolveMissingItems
	
	function ItemBuff:ResolveBuffItemId(buff, setLink)
		if buff.found then
			return true
		end

		local linked, changed
		local itemName = GetItemInfo(buff.itemid)
		
		if itemName then
			
			if missingItems and missingItems[buff] then
				missingItems[buff] = nil
				if not next(missingItems) then
					missingItems = del(missingItems)
				end
			end
			
			buff.found = true
			
			if itemName ~= buff.name then
				ItemBuff:Debug('%s solved as %s', buff.name, itemName)
				buff:ChangeName(itemName)
				changed = true
			end
			
		elseif setLink then
			ItemBuff:Debug('Linking #%d', buff.itemid)			
			AceLibrary('Gratuity-2.0'):SetHyperlink('item:'..buff.itemid)
			linked = true
			
		else
			ItemBuff:Debug('Register #%d for scanning', buff.itemid)			
			if not missingItems then
				missingItems = new()
			end
			missingItems[buff] = true			
		end
		
		return buff.found, linked, changed
		
	end
	
	function ItemBuff:SetupSecureButton(buff, button)
		self:Debug('Setup using for item %q', buff.name)
		button:SetAttribute('type', 'item')
		button:SetAttribute('item', buff.name)
	end
	
end

-------------------------------------------------------------------------------
-- Buffs from spells
-------------------------------------------------------------------------------

do
	local SpellBuff = BuffSourceClass:new(SOURCE_SPELL, IsUsableSpell, GetSpellCooldown)

	function SpellBuff:OnInitialize()
		if next(self.allBuffs) then
			self:ScanSpells()
		end
	end
		
	function SpellBuff:OnEnable()
		if next(self.allBuffs) then
			self:RegisterEvent('SPELLS_CHANGED', 'ScanSpells')
			self:ScanSpells()
		end
	end
	
	function SpellBuff:ScanSpells()
		local dirty = false
		local knownSpells = new()
		
		local i = 1
		while true do
			local name = GetSpellName(i, BOOKTYPE_SPELL)
			if not name then
				break
			end
			knownSpells[name] = true
			i = i + 1
		end
		
		for buff in pairs(self.allBuffs) do
			local found = knownSpells[buff.name]
			if found and not buff.found then
				-- SimpleSelfRebuff:Debug('New spell: %q', buff.name)
				buff.found = true
				dirty = true
			elseif not found and buff.found then
				-- SimpleSelfRebuff:Debug('Old spell: %q', buff.name)
				buff.found = nil
				dirty = true
			end
		end
		
		knownSpells = del(knownSpells)
	
		if dirty then
			self:TriggerEvent('SimpleSelfRebuff_AvailableBuffsChanged')
		end	
	end
	
	function SpellBuff:SetupSecureButton(buff, button)
		self:Debug('Setup casting for spell %q', buff.name)
		button:SetAttribute('type', 'spell')
		button:SetAttribute('spell', buff.name)
	end
	
end

-------------------------------------------------------------------------------
-- Auras
-------------------------------------------------------------------------------

do
	local AuraBuff = BuffTargetClass:new(TARGET_AURA)
	
	function AuraBuff:OnEnable()
		if next(self.allBuffs) then
			self:RegisterEvent('PLAYER_AURAS_CHANGED', 'ScanAuras')
			self:RegisterEvent('SimpleSelfRebuff_BuffSetupChanged', 'ScanAuras')
			self:ScanAuras()
		end
	end
	
	function AuraBuff:ScanAuras()
		local hasTimedBuff
		local seen = new()
		
		for i = 1, 32 do
			-- Only take in account buffs we actually cast
			local buffName, _, _, _, duration, timeLeft = UnitBuff('player', i)	
			if not buffName then
				break
			end
			local buff = buffName and self.buffs[buffName]
			if buff then
				if timeLeft and timeLeft > 0 then
					hasTimedBuff = true
				end
				seen[buff] = true
				if buff:SetAsActualBuff(true, (duration and true or false), timeLeft) then
					self:Debug('Found aura %s (%q)', buff.name, timeLeft)
				end
			end
		end
		
		for buff in pairs(self.allBuffs) do
			if not seen[buff] and buff:IsActualBuff() then
				self:Debug('%s faded out', buff.name)
				buff:SetAsActualBuff(false)
			end
		end
		
		seen = del(seen)
		
		if hasTimedBuff then
			self:ScheduleEvent('SimpleSelfRebuff_ScanAuras', self.ScanAuras, 5, self)
		end
	end
	
	function AuraBuff:DumpDiagnostic()
		for i = 1, 32 do
			-- Only take in account buffs we actually cast
			local buffName, _, _, _, duration, timeLeft = UnitBuff('player', i)	
			if not buffName then
				break
			end
			if buffName and self.buffs[buffName] then
				SimpleSelfRebuff:Print("  buff=%q, timeLeft=%q", buffName, timeLeft)
			end
		end		
	end
	
	function AuraBuff:SetupSecureButton(buff, button)
		button:SetAttribute('unit', 'player')
	end
	
end

-------------------------------------------------------------------------------
-- Tracking
-------------------------------------------------------------------------------

do
	local TrackingBuff = BuffTargetClass:new(TARGET_TRACKING)
		
	local textures = {}
	
	function TrackingBuff:OnBuffRegister(buff)
		local texture = BS:GetSpellIcon(buff.name)
		if not texture then
			error('Unknown tracking texture for spell %q', buff.name)
		elseif textures[texture] then
			error('Tracking texture already set for %q: %q', buff.name, self.textures[texture].name)
		end
		textures[texture] = buff
	end
	
	function TrackingBuff:OnEnable()
		self:RegisterEvent('PLAYER_AURAS_CHANGED', 'ScanTracking')
		self:RegisterEvent('SimpleSelfRebuff_BuffSetupChanged', 'ScanTracking')
		self:ScanTracking()
	end
	
	function TrackingBuff:ScanTracking()
		local texture = GetTrackingTexture()
		local buff = texture and textures[texture]
		if buff then
			buff:SetAsActualBuff(true, true)
		else
			for buff in pairs(self.allBuffs) do
				if buff:IsActualBuff() then
					buff:SetAsActualBuff(false)
				end
			end
		end
	end
	
	function TrackingBuff:SetupSecureButton(buff, button)
		button:SetAttribute('unit', 'player')
	end

end

-------------------------------------------------------------------------------
-- Weapon enchantment buff
-------------------------------------------------------------------------------

do
	local WeaponBuffClass = Class(BuffTargetClass)

	function WeaponBuffClass.prototype:init(target, slotName, infoIndex, validEquipLocs)
		local slot = GetInventorySlotInfo(slotName)
		if type(slot) ~= 'number' then
			error("Invalid slot name %q => %q", slotName, slot)
		end
		
		WeaponBuffClass.super.prototype.init(self, target)
		
		self.slot = slot
		self.infoIndex = infoIndex
		self.validEquipLocs = validEquipLocs
	end
	
	function WeaponBuffClass.prototype:OnInitialize() 
		if next(self.allBuffs) then
			self:CheckSlot()
		end
	end
	
	function WeaponBuffClass.prototype:OnEnable() 
		if next(self.allBuffs) then
			self:RegisterBucketEvent('UNIT_INVENTORY_CHANGED', 0.5, 'CheckSlot')
			self:ScheduleRepeatingEvent('SimpleSelfRebuff_ScanWeaponBuffs', self.ScanWeaponBuffs, 1, self)
			self:RegisterEvent('SimpleSelfRebuff_BuffSetupChanged', 'ScanWeaponBuffs')
			self:CheckSlot()
			self:ScanWeaponBuffs()
		end	
	end
		
	function WeaponBuffClass.prototype:ScanWeaponBuffs()
		local hasEnchant, timeLeft = select(self.infoIndex, GetWeaponEnchantInfo())
		local category = next(self.allBuffs).category		
		category:SetActualBuff(
			hasEnchant and category:GetExpectedBuff(), 
			hasEnchant, 
			hasEnchant and timeLeft and math.floor(timeLeft/1000)
		)
	end
		
	function WeaponBuffClass.prototype:CheckSlot()
		local itemLink = GetInventoryItemLink("player", self.slot)
		local itemEquipLoc = itemLink and select(9, GetItemInfo(itemLink))
		if itemEquipLoc ~= self.itemEquipLoc then
			self.itemEquipLoc = itemEquipLoc
			self:TriggerEvent('SimpleSelfRebuff_AvailableBuffsChanged')
		end
	end		
		
	function WeaponBuffClass.prototype:SetupSecureButton(buff, button)
		self:Debug('Setting up for buff %q on %q', buff.name, self.slot)
		button:SetAttribute('target-slot', self.slot)
	end
		
	function WeaponBuffClass.prototype:IsTargetValid(buff)
		return self.itemEquipLoc and self.validEquipLocs[self.itemEquipLoc]
	end
	
	function WeaponBuffClass.prototype:DumpDiagnostic()
		SimpleSelfRebuff:Print("  Slot: %d, itemLoc: %s", self.slot, self.itemEquipLoc)
		local hasEnchant, timeLeft, charges = select(self.infoIndex, GetWeaponEnchantInfo())
		SimpleSelfRebuff:Print("  WeaponInfo: hasEnchant=%q, timeLeft=%q, charges=%q", hasEnchant, timeLeft, charges)
	end
	
	WeaponBuffClass:new(
		TARGET_MAINHAND, 
		'MainHandSlot',
		1,
		{
			INVTYPE_WEAPON         = true,
			INVTYPE_2HWEAPON       = true,
			INVTYPE_WEAPONMAINHAND = true,
		}
	)

	WeaponBuffClass:new(
		TARGET_OFFHAND,
		'SecondaryHandSlot',
		4,
		{
			INVTYPE_WEAPON         = true,
			INVTYPE_WEAPONOFFHAND  = true,
		}
	)

end

-------------------------------------------------------------------------------
-- Buff registry
-------------------------------------------------------------------------------

do
	local setupFuncs = {}

	function SimpleSelfRebuff:RegisterBuffSetup(func)
		if type(func) == "function" then
			tinsert(setupFuncs, func)
		else
			error("Arg #1 to RegisterBuffs must be a function, not %q", type(func))
		end
	end
	
	function SimpleSelfRebuff:ProcessRegistry()
		if not setupFuncs then
			return
		end

		-- Babble-Spell isn't needed before we reach this point
		BS = AceLibrary("Babble-Spell-2.2")
		
		-- Prebuild categories
		self:GetCategory(self.CATEGORY_TRACKING, self.TARGET_TRACKING)
		self:GetCategory(self.CATEGORY_MAINHAND, self.TARGET_MAINHAND)
		self:GetCategory(self.CATEGORY_OFFHAND,  self.TARGET_OFFHAND)
		
		-- Load all funcs
		local selfClassName = select(2, UnitClass('player'))
		for i,func in ipairs(setupFuncs) do
			local succes, msg = pcall(func, self, BS, selfClassName, L)
			if not succes then
				geterrorhandler()(("Error in setupFunc #%d: %s"):format(i, msg))
			end
		end
		
		-- Clean up
		setupFuncs = del(setupFuncs)
		
		-- Purge configuration
		for categoryName in pairs(db_char.categories) do
			if not categories[categoryName] then
				db_char.categories[categoryName] = nil
			end
		end
		
		self:TriggerEvent('SimpleSelfRebuff_RegistryProcessed')
		
		self:UpdateCategories()		
		self:BuildOptions()
	end
end
	
function SimpleSelfRebuff:GetCategory(name, ...)
	if type(name) ~= 'string' then
		error("Argument #2 to GetCategory should be a string, not %q", type(name))
	end
	if not categories[name] then
		categories[name] = CategoryClass:new(name, ...)
	end
	return categories[name]
end

function SimpleSelfRebuff:AddStandaloneBuff(name, ...)
	if type(name) ~= 'string' then
		error("Argument #2 to AddStandaloneBuff should be a string, not %q", type(name))
	end
	self:GetCategory(name):add(name, ...)
end

function SimpleSelfRebuff:AddMultiStandaloneBuffs(...)
	for i=1,select('#', ...) do
		local name = select(i, ...)
		self:AddStandaloneBuff(name)
	end
end
